import uproot
from atuproot.BEvents import BEvents
from .masked_tree import MaskedUprootTree


class EventRanger():
    def __init__(self):
        self._owner = None

    def set_owner(self, owner):
        self._owner = owner

    @property
    def start_entry(self):
        return (self._owner.start_block + self._owner.iblock) * self._owner.nevents_per_block

    @property
    def stop_entry(self):
        i_block = min(self._owner.iblock + 1, self._owner.nblocks)
        stop_entry = (self._owner.start_block + i_block) * self._owner.nevents_per_block
        return min(self._owner.nevents_in_tree, stop_entry)

    @property
    def entries_in_block(self):
        if self._owner and self._owner.iblock > -1:
            return self.stop_entry - self.start_entry
        return None


class BEventsWrapped(BEvents):
    def __init__(self, tree, *args, **kwargs):
        ranges = EventRanger()
        tree = MaskedUprootTree(tree, ranges)
        super(BEventsWrapped, self).__init__(tree, *args, **kwargs)
        ranges.set_owner(self)

    def _block_changed(self):
        self.tree.reset_mask()
        self.tree.reset_cache()

    def __getitem__(self, i):
        result = super(BEventsWrapped, self).__getitem__(self, i)
        self._block_changed()
        return result

    def __iter__(self):
        for value in super(BEventsWrapped, self).__iter__():
            self._block_changed()
            yield value
        self._block_changed()


class EventBuilder(object):
    def __init__(self, config):
        self.config = config

    def __repr__(self):
        return '{}({!r})'.format(
            self.__class__.__name__,
            self.config,
        )

    def __call__(self):
        if len(self.config.inputPaths) != 1:
            # TODO - support multiple inputPaths
            raise AttributeError("Multiple inputPaths not yet supported")

        # Try to open the tree - some machines have configured limitations
        # which prevent memmaps from begin created. Use a fallback - the
        # localsource option
        try:
            rootfile = uproot.open(self.config.inputPaths[0])
            tree = rootfile[self.config.treeName]
        except MemoryError:
            rootfile = uproot.open(self.config.inputPaths[0],
                                   localsource=uproot.FileSource.defaults)
            tree = rootfile[self.config.treeName]

        events = BEventsWrapped(tree,
                                self.config.nevents_per_block,
                                self.config.start_block,
                                self.config.stop_block)
        events.config = self.config
        return events



import six
import numpy as np
import pandas as pd
from ..expressions import evaluate
from ..define.reductions import get_awkward_reduction


def safe_and(left, right):
    if left is None:
        return right
    if right is None:
        return left
    return left & right


def safe_or(left, right):
    if left is None:
        return right
    if right is None:
        return left
    return left | right


class Counter():
    def __init__(self, weights):
        self._weights = weights
        self._w_counts = np.zeros(len(weights))
        self._counts = 0

    @staticmethod
    def get_unweighted_increment(data, mask):
        if mask is None:
            return len(data)
        elif mask.dtype.kind == "b":
            return np.count_nonzero(mask)
        else:
            return len(mask)

    @staticmethod
    def get_weighted_increment(weight_names, data, mask):
        weights = data.arrays(weight_names, outputtype=lambda *args: np.array(args))
        if mask is not None:
            weights = weights[:, mask]
        return weights.sum(axis=1)

    def increment(self, data, is_mc, mask=None):
        unweighted_increment = self.get_unweighted_increment(data, mask)
        self._counts += unweighted_increment

        if not self._weights:
            return

        if not is_mc:
            self._w_counts += unweighted_increment
            return

        weighted_increments = self.get_weighted_increment(self._weights, data, mask)
        self._w_counts += weighted_increments

    @property
    def counts(self):
        return (self._counts,) + tuple(self._w_counts)

    def add(self, rhs):
        self._w_counts += rhs._w_counts
        self._counts += rhs._counts


class BaseFilter(object):

    def __init__(self, selection, depth, cut_id, weights):
        self._unique_id = ",".join(map(str, cut_id))

        self.selection = selection
        self.depth = depth
        self.passed_excl = Counter(weights)
        self.totals_incl = Counter(weights)
        self.passed_incl = Counter(weights)
        self.weights = weights

    @property
    def index_values(self):
        output = [(self._unique_id, self.depth, str(self))]
        if isinstance(self.selection, list):
            output = sum([sel.index_values for sel in self.selection], output)
        return output

    @property
    def values(self):
        output = [self.passed_excl.counts + self.passed_incl.counts + self.totals_incl.counts]
        if isinstance(self.selection, list):
            output += sum([sel.values for sel in self.selection], [])
        return output

    @property
    def columns(self):
        nweights = len(self.weights) + 1
        row1 = ["passed_only_cut"] * nweights
        row1 += ["passed_incl"] * nweights
        row1 += ["totals_incl"] * nweights
        row2 = (["unweighted"] + self.weights) * 3
        return [row1, row2]

    def to_dataframe(self):
        index_names = ("unique_id", "depth", "cut")
        index = pd.MultiIndex.from_tuples(self.index_values, names=index_names)
        columns = pd.MultiIndex.from_arrays(self.columns)
        return pd.DataFrame(self.values, columns=columns, index=index)

    def merge(self, rhs):
        self.totals_incl.add(rhs.totals_incl)
        self.passed_incl.add(rhs.passed_incl)
        self.passed_excl.add(rhs.passed_excl)
        if isinstance(self.selection, list):
            for sub_lhs, sub_rhs in zip(self.selection, rhs.selection):
                sub_lhs.merge(sub_rhs)
        return self

    def increment_counters(self, data, is_mc, excl, before, after):
        self.passed_excl.increment(data, is_mc, excl)
        self.passed_incl.increment(data, is_mc, after)
        self.totals_incl.increment(data, is_mc, before)

    def __repr__(self):
        rep = ": {!r}"
        if isinstance(self.selection, list):
            rep = ": [{!r}]"
        rep = self.__class__.__name__ + rep.format(self.selection)
        return rep


class ReduceSingleCut(BaseFilter):
    def __init__(self, stage_name, depth, cut_id, weights, selection):
        super(ReduceSingleCut, self).__init__(selection, depth, cut_id, weights)
        self._str = str(selection)
        self.reduction = get_awkward_reduction(stage_name,
                                               selection.get("reduce"),
                                               fill_missing=False)
        self.formula = selection.get("formula")

    def __call__(self, data, is_mc, **kwargs):
        mask = evaluate(data, self.formula)
        mask = self.reduction(mask)
        return mask

    def __str__(self):
        return self._str


class SingleCut(BaseFilter):
    def __call__(self, data, is_mc, **kwargs):
        mask = evaluate(data, self.selection)
        return mask

    def __str__(self):
        return self.selection


class All(BaseFilter):
    def __call__(self, data, is_mc,
                 current_mask=None, combine_op=safe_and):
        mask = np.ones(len(data), dtype=bool)
        for sel in self.selection:
            excl_mask = sel(data, is_mc,
                            current_mask=combine_op(current_mask, mask),
                            combine_op=safe_and)
            new_mask = mask & excl_mask
            sel.increment_counters(data, is_mc, excl=excl_mask,
                                   after=new_mask, before=mask)
            mask = new_mask
        return mask

    def __str__(self):
        return "All"


class Any(BaseFilter):
    def __call__(self, data, is_mc,
                 current_mask=None, combine_op=safe_or):
        mask = np.zeros(len(data), dtype=bool)
        for sel in self.selection:
            excl_mask = sel(data, is_mc,
                            current_mask=current_mask,
                            combine_op=combine_op)
            new_mask = mask | excl_mask
            sel.increment_counters(data, is_mc, excl=excl_mask,
                                   after=combine_op(new_mask, current_mask),
                                   before=current_mask)
            mask = new_mask
        return mask

    def __str__(self):
        return "Any"


class OuterCounterIncrementer(BaseFilter):
    def __init__(self, *args, **kwargs):
        super(OuterCounterIncrementer, self).__init__(*args, **kwargs)
        self._wrapped_selection = BaseFilter.__getattribute__(self, "selection")

    def __call__(self, data, is_mc):
        mask = self._wrapped_selection(data, is_mc)
        self._wrapped_selection.increment_counters(data, is_mc, excl=mask, after=mask, before=None)
        return mask

    def __getattribute__(self, name):
        if name in ["__call__", "_wrapped_selection"]:
            return BaseFilter.__getattribute__(self, name)
        return BaseFilter.__getattribute__(self, "selection").__getattribute__(name)


def build_selection(stage_name, config, weights=[]):
    """Creates event selectors based on the configuration.

    Parameters:
        stage_name: Used to help in error messages.
        config: The event selection configuration.
        weights: How to weight events, used to produce the resulting cut
            efficiency table.

    Raises:
        RuntimeError: if any of the configurations are invalid.
    """
    selection = handle_config(stage_name, config, weights)
    return OuterCounterIncrementer(selection, depth=-1, cut_id=[-1], weights=weights)


def handle_config(stage_name, config, weights, depth=0, cut_id=[0]):
    if isinstance(config, six.string_types):
        return SingleCut(config, depth, cut_id, weights)
    if not isinstance(config, dict):
        raise RuntimeError(stage_name + ": Selection config not a dict")
    if len(config) == 2:
        return ReduceSingleCut(stage_name, depth, cut_id, weights, config)
    elif len(config) != 1:
        raise RuntimeError(stage_name + ":Selection config has too many keys")

    method, in_selections = tuple(config.items())[0]
    if method not in ("All", "Any"):
        raise RuntimeError(stage_name + ": Unknown selection combination method," + method)

    selections = []
    for i, sel in enumerate(in_selections):
        cut = handle_config(stage_name, sel, weights, depth + 1, cut_id=cut_id + [i])
        selections.append(cut)
    if method == "All":
        return All(selections, depth, cut_id, weights)
    if method == "Any":
        return Any(selections, depth, cut_id, weights)



import pandas as pd
import numpy as np
from .tree_wrapper import WrappedTree


class MaskedUprootTree(object):
    def __init__(self, tree, event_ranger, mask=None):
        if isinstance(tree, MaskedUprootTree):
            self.tree = tree.tree
            self._mask = tree._mask
            return

        self.tree = WrappedTree(tree, event_ranger)
        self.event_ranger = event_ranger

        if mask is None:
            self._mask = None
            return

        self._mask = _normalise_mask(mask, len(self.tree))

    class PandasWrap():
        def __init__(self, owner):
            self._owner = owner

        def df(self, *args, **kwargs):
            df = self._owner.tree.pandas.df(*args, **kwargs)
            if self._owner._mask is None:
                return df
            masked = mask_df(df, self._owner._mask, self._owner.event_ranger.start_entry)
            return masked

    @property
    def pandas(self):
        return MaskedUprootTree.PandasWrap(self)

    def unmasked_array(self, *args, **kwargs):
        return self.tree.array(*args, **kwargs)

    def unmasked_arrays(self, *args, **kwargs):
        return self.tree.arrays(*args, **kwargs)

    def array(self, *args, **kwargs):
        array = self.tree.array(*args, **kwargs)
        if self._mask is None:
            return array
        return array[self._mask]

    def arrays(self, *args, **kwargs):
        arrays = self.tree.arrays(*args, **kwargs)
        if self._mask is None:
            return arrays
        if isinstance(arrays, dict):
            return {k: v[self._mask] for k, v in arrays.items()}
        if isinstance(arrays, tuple):
            return tuple([v[self._mask] for v in arrays])
        if isinstance(arrays, list):
            return [v[self._mask] for v in arrays]
        if isinstance(arrays, pd.DataFrame):
            return mask_df(arrays, self._mask, self.event_ranger.start_entry)
        if isinstance(arrays, np.ndarray):
            if arrays.ndim == 1:
                return arrays[self._mask]
            if arrays.ndim == 2:
                if arrays.shape[1] == len(self.tree):
                    return arrays[:, self._mask]
            msg = "Unexpected numpy array for mask, shape:%s, mask length: %s"
            raise NotImplementedError(msg % (arrays.shape, len(self)))
        return arrays[self._mask]

    @property
    def mask(self):
        return self._mask

    def apply_mask(self, new_mask):
        if self._mask is None:
            self._mask = _normalise_mask(new_mask, len(self.tree))
        else:
            self._mask = self._mask[new_mask]

    def __len__(self):
        if self._mask is None:
            return len(self.tree)
        return len(self._mask)

    def __contains__(self, element):
        return self.tree.__contains__(element)

    def __getattr__(self, attr):
        return getattr(self.tree, attr)

    def reset_mask(self):
        self._mask = None


def _normalise_mask(mask, tree_length):
    if isinstance(mask, (tuple, list)):
        mask = np.array(mask)
    elif not isinstance(mask, np.ndarray):
        raise RuntimeError("mask is not a numpy array, a list, or a tuple")

    if np.issubdtype(mask.dtype, np.integer):
        return mask
    elif mask.dtype.kind == "b":
        if len(mask) != tree_length:
            raise RuntimeError("boolean mask has a different length to the input tree")
        return np.where(mask)[0]


def mask_df(df, mask, start_event):
    mask = mask + start_event

    # Either of these methods could work on a general df (multiindex, or not) but they
    # have opposite performances, so check if multi-index and choose accordingly
    if isinstance(df.index, pd.MultiIndex):
        broadcast_map = np.isin(df.index.get_level_values("entry"), mask)
        masked = df.iloc[broadcast_map]
    else:
        masked = df.loc[mask]
    return masked



try:
    import aghast
    has_aghast = True
except ImportError as ex:
    if "aghast" not in str(ex):
        raise

    class AghastCatcher:
        def __getattr__(self, attr):
            msg = "Aghast is not installed but has been needed."
            msg += "\nInstall it using conda:"
            msg += "\n\n       conda install aghast"
            raise ImportError(msg)

    aghast = AghastCatcher()
    has_aghast = False



import os
import sys


if sys.version_info < (3, 2):
    import errno

    def mkdir_p(path):
        try:
            os.makedirs(path)
        except OSError as exc:  # Python >2.5
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                raise
else:
    def mkdir_p(path):
        return os.makedirs(path, exist_ok=True)



import numpy as np
import six


__all__ = ["get_pandas_reduction"]


class BadReductionConfig(Exception):
    pass


class JaggedNth(object):
    def __init__(self, index, fill_missing, force_float=True):
        self.index = index
        self.fill_missing = fill_missing
        self.dtype = None
        if force_float and isinstance(fill_missing, int):
            if fill_missing is True or fill_missing is False:
                self.dtype = bool
            else:
                self.dtype = float

    def __call__(self, array):
        mask = array.counts > abs(self.index) - int(self.index < 0)
        output = np.full(len(array), self.fill_missing, dtype=self.dtype)
        output[mask] = array[mask, self.index]
        return output


class JaggedMethod(object):
    def __init__(self, method):
        self.method_name = method

    def __call__(self, array):
        return getattr(array, self.method_name)()


class JaggedProperty(object):
    def __init__(self, prop_name):
        self.prop_name = prop_name

    def __call__(self, array):
        return getattr(array, self.prop_name)


_jagged_methods = ["sum", "prod", "any", "all", "count_nonzero",
                   "max", "min", "argmin", "argmax"]
_jagged_properties = ["counts"]


def get_awkward_reduction(stage_name, reduction, fill_missing=np.nan):
    if isinstance(reduction, six.integer_types):
        return JaggedNth(int(reduction), fill_missing)

    if not isinstance(reduction, six.string_types):
        msg = "{}: requested reduce method is not a string or an int"
        raise BadReductionConfig(msg.format(stage_name))

    if reduction in _jagged_methods:
        return JaggedMethod(reduction)
    if reduction in _jagged_properties:
        return JaggedProperty(reduction)

    msg = "{}: Unknown method to reduce: '{}'"
    raise BadReductionConfig(msg.format(stage_name, reduction))


_pandas_aggregates = ["sum", "prod", "max", "min", "argmax", "argmin"]
_numpy_ops = ["count_zero"]


class PandasAggregate(object):
    def __init__(self, method):
        self.method = method

    def __call__(self, groups):
        return groups.agg(self.method)


class PandasNth(object):
    def __init__(self, index):
        self.index = index

    def __call__(self, groups):
        return groups.nth(self.index)


def get_pandas_reduction(stage_name, reduction):
    if not isinstance(reduction, (six.string_types, six.integer_types)):
        msg = "{}: requested reduce method is not a string or an int"
        raise BadReductionConfig(msg.format(stage_name))

    if reduction in _pandas_aggregates:
        return PandasAggregate(reduction)
    elif reduction in _numpy_ops:
        op = getattr(np, reduction)
        return PandasAggregate(op)
    else:
        try:
            index = int(reduction)
        except ValueError:
            pass
        else:
            return PandasNth(index)

    msg = "{}: Unknown method to reduce: '{}'"
    raise BadReductionConfig(msg.format(stage_name, reduction))



import numpy as np
import re
import numexpr
import tokenize
import awkward
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


__all__ = ["get_branches", "evaluate"]


constants = {"nan": np.nan,
             "inf": np.inf,
             "pi": np.pi,
             "e": np.e,
             }


def get_branches(cut, valid):
    valid = [v.decode("utf-8") for v in valid]

    branches = []
    string = StringIO(cut).readline
    tokens = tokenize.generate_tokens(string)
    current_branch = []
    for toknum, tokval, _, _, _ in tokens:
        if toknum == tokenize.NAME:
            if ".".join(current_branch + [tokval]) in valid:
                current_branch.append(tokval)
                continue
        if tokval == ".":
            continue
        if current_branch:
            branches.append(".".join(current_branch))
            current_branch = []
    return branches


def deconstruct_jaggedness(array, counts):
    if not isinstance(array, awkward.array.base.AwkwardArrayWithContent):
        return array, counts

    array = array.compact()
    counts.insert(0, array.counts)
    return deconstruct_jaggedness(array.content, counts)


def reconstruct_jaggedness(array, counts):
    for count in counts:
        array = awkward.JaggedArray.fromcounts(count, array)
    return array


class TreeToDictAdaptor():
    """
    Make an uproot tree look like a dict for numexpr
    """
    def __init__(self, tree, alias_dict):
        self.tree = tree
        self.counts = None
        self.aliases = alias_dict

    def __getitem__(self, item):
        if item in constants:
            return constants[item]
        full_item = self.aliases.get(item, item)
        array = self.tree.array(full_item)
        array = self.strip_jaggedness(array)
        return array

    def __contains__(self, item):
        return item in self.tree or item in self.aliases

    def __iter__(self):
        for i in self.tree:
            yield i

    def strip_jaggedness(self, array):
        array, new_counts = deconstruct_jaggedness(array, counts=[])
        if self.counts is not None:
            if not all(np.array_equal(c, n) for c, n in zip(self.counts, new_counts)):
                raise RuntimeError("Operation using arrays with different jaggedness")
        else:
            self.counts = new_counts
        return array

    def apply_jaggedness(self, array):
        if self.counts is None:
            return array
        result = reconstruct_jaggedness(array, self.counts)
        return result


attribute_re = re.compile(r"([a-zA-Z]\w*)\s*\.\s*(\w+)")


def preprocess_expression(expression):
    alias_dict = {}
    replace_dict = {}
    for match in attribute_re.finditer(expression):
        original = match.group(0)
        alias = match.expand(r"\1__DOT__\2")
        alias_dict[alias] = original
        replace_dict[original] = alias
    clean_expr = attribute_re.sub(lambda x: replace_dict[x.group(0)], expression)
    return clean_expr, alias_dict


def evaluate(tree, expression):
    cleaned_expression, alias_dict = preprocess_expression(expression)
    adaptor = TreeToDictAdaptor(tree, alias_dict)
    result = numexpr.evaluate(cleaned_expression, local_dict=adaptor)
    result = adaptor.apply_jaggedness(result)
    return result
